import os
import shutil
from argparse import Namespace, _SubParsersAction
from typing import List

from InquirerPy import prompt

from ..model_config import MODEL_TYPES
from ..repo_build import RepositoryBuilder
from ._utils import list_model_upload_examples
from .base import BaseClarifaiCli


class CreateCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: _SubParsersAction):
    creator_parser = parser.add_parser("create", help="Create component of Clarifai platform")
    sub_creator_parser = creator_parser.add_subparsers()

    SubCreateModelCli.register(sub_creator_parser)

    creator_parser.set_defaults(func=CreateCli)


class SubCreateModelCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: _SubParsersAction):
    model_parser = parser.add_parser("model")
    model_parser.add_argument(
        "--working-dir",
        type=str,
        required=True,
        help="Path to your working dir. Create new dir if it does not exist")
    model_parser.add_argument(
        "--from-example",
        required=False,
        action="store_true",
        help="Create repository from example")
    model_parser.add_argument(
        "--example-id",
        required=False,
        type=str,
        help="Example id, run `clarifai example list` to list of examples")

    model_parser.add_argument(
        "--type",
        type=str,
        choices=MODEL_TYPES,
        required=False,
        help="Clarifai supported model types.")
    model_parser.add_argument(
        "--image-shape",
        nargs='+',
        type=int,
        required=False,
        help="H W dims for models with an image input type. H and W each have a max value of 1024",
        default=[-1, -1])
    model_parser.add_argument(
        "--max-bs", type=int, default=1, required=False, help="Max batch size")

    model_parser.add_argument(
        "--overwrite", action="store_true", help="Overwrite working-dir if exists")

    model_parser.set_defaults(func=SubCreateModelCli)

  def __init__(self, args: Namespace) -> None:
    self.working_dir: str = args.working_dir
    self.from_example = args.from_example
    self.example_id = args.example_id
    self.overwrite = args.overwrite

    if os.path.exists(self.working_dir):
      if self.overwrite:
        print(f"Overwrite {self.working_dir}")
      else:
        raise FileExistsError(
            f"{self.working_dir} exists. If you want to overwrite it, please set `--overwrite` flag"
        )

    # prevent wrong args when creating from example
    if not self.from_example:
      self.image_shape: List[int] = args.image_shape

      self.type: str = args.type
      self.max_bs: int = args.max_bs

    else:
      if not self.example_id:
        questions = [
            {
                "type": "list",
                "message": "Select an example:",
                "choices": list_model_upload_examples(),
            },
        ]
        result = prompt(questions)
        self.example_id = result[0]

      else:
        available_examples = list(list_model_upload_examples().keys())
        assert self.example_id in available_examples, f"Available examples are: {available_examples}, got {self.example_id}."

  def run(self):
    if self.from_example:
      os.makedirs(self.working_dir, exist_ok=True)
      model_repo, readme = list_model_upload_examples()[self.example_id]
      shutil.copytree(model_repo, self.working_dir, dirs_exist_ok=True)
      if readme:
        shutil.copy(readme, os.path.join(self.working_dir, "readme.md"))

    else:
      RepositoryBuilder.init_repository(
          self.type,
          self.working_dir,
          backend="triton",
          max_batch_size=self.max_bs,
          image_shape=self.image_shape)

    from itertools import islice
    from pathlib import Path

    def tree(dir_path: Path,
             level: int = -1,
             limit_to_directories: bool = False,
             length_limit: int = 1000):
      # prefix components:
      space = '    '
      branch = '│   '
      # pointers:
      tee = '├── '
      last = '└── '
      """Given a directory Path object print a visual tree structure"""
      dir_path = Path(dir_path)  # accept string coerceable to Path
      files = 0
      directories = 0

      def inner(dir_path: Path, prefix: str = '', level=-1):
        nonlocal files, directories
        if not level:
          return  # 0, stop iterating
        if limit_to_directories:
          contents = [d for d in dir_path.iterdir() if d.is_dir()]
        else:
          contents = list(dir_path.iterdir())
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
          if path.is_dir():
            yield prefix + pointer + path.name
            directories += 1
            extension = branch if pointer == tee else space
            yield from inner(path, prefix=prefix + extension, level=level - 1)
          elif not limit_to_directories:
            yield prefix + pointer + path.name
            files += 1

      print(dir_path.name)
      iterator = inner(dir_path, level=level)
      for line in islice(iterator, length_limit):
        print(line)
      if next(iterator, None):
        print(f'... length_limit, {length_limit}, reached, counted:')
      print(f'\n{directories} directories' + (f', {files} files' if files else ''))

    print("-" * 75)
    print(f"* Created repository at: {self.working_dir}")
    tree(self.working_dir)
    print()
    print("* Please make sure your code is tested using `test.py` before uploading")
    print("-" * 75)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Triton inference server Python Backend Model."""

import os
import sys

try:
  import triton_python_backend_utils as pb_utils
except ModuleNotFoundError:
  pass
from clarifai.models.model_serving.model_config.inference_parameter import parse_req_parameters


class TritonPythonModel:
  """
  Triton Python BE Model.
  """

  def initialize(self, args):
    """
    Triton server init.
    """
    sys.path.append(os.path.dirname(__file__))
    from inference import InferenceModel

    self.inference_obj = InferenceModel()

    # Read input_name from config file
    self.input_names = [inp.name for inp in self.inference_obj.config.serving_backend.triton.input]

  def execute(self, requests):
    """
    Serve model inference requests.
    """
    responses = []

    for request in requests:
      try:
        parameters = request.parameters()
      except Exception:
        print(
            "It seems this triton version does not support `parameters()` in request. "
            "Please upgrade tritonserver version otherwise can not use `inference_parameters`. Error message: {e}"
        )
        parameters = None

      parameters = parse_req_parameters(parameters) if parameters else {}

      if len(self.input_names) == 1:
        in_batch = pb_utils.get_input_tensor_by_name(request, self.input_names[0])
        in_batch = in_batch.as_numpy()
        data = in_batch
      else:
        data = {}
        for input_name in self.input_names:
          in_batch = pb_utils.get_input_tensor_by_name(request, input_name)
          in_batch = in_batch.as_numpy() if in_batch is not None else []
          data.update({input_name: in_batch})

      inference_response = self.inference_obj._tritonserver_predict(
          input_data=data, inference_parameters=parameters)
      responses.append(inference_response)

    return responses



import logging
from copy import deepcopy
from dataclasses import asdict, dataclass, field
from typing import Any, List

import yaml

from .inference_parameter import InferParam
from .output import *  # noqa: F403
from .triton import DType  # noqa
from .triton import Device, DynamicBatching, InputConfig, OutputConfig, TritonModelConfig

logger = logging.getLogger(__name__)

__all__ = ["ModelTypes", "ModelConfigClass", "MODEL_TYPES", "get_model_config", "load_user_config"]


# Clarifai model types
@dataclass
class ModelTypes:
  """ All supported Clarifai model type names
  """
  visual_detector: str = "visual-detector"
  visual_classifier: str = "visual-classifier"
  text_classifier: str = "text-classifier"
  text_to_text: str = "text-to-text"
  text_embedder: str = "text-embedder"
  text_to_image: str = "text-to-image"
  visual_embedder: str = "visual-embedder"
  visual_segmenter: str = "visual-segmenter"
  multimodal_embedder: str = "multimodal-embedder"

  def __post_init__(self):
    self.all = list(asdict(self).values())

  @property
  def image_input_models(self):
    """ Return list of model types having image as input or one of inputs"""
    _visual = [each for each in self.all if each.startswith("visual")]

    return _visual + [self.multimodal_embedder]


@dataclass
class FieldMapsConfig:
  input_fields_map: dict = field(default_factory=dict)
  output_fields_map: dict = field(default_factory=dict)


@dataclass
class ServingBackendConfig:
  """
  """
  triton: TritonModelConfig = None


@dataclass
class ClarifaiModelConfig:
  """Clarifai necessary configs for building/uploading/creation

  Args:
    field_maps (FieldMapsConfig): Field maps config
    output_type (dataclass): model output type
    labels (List[str]): list of concept names
    inference_parameters (List[InferParam]): list of inference parameters
    clarifai_model_id (str): Clarifai model id on the platform
    type (str): one of `MODEL_TYPES`
    clarifai_user_app_id (str): User ID and App ID separated by '/', e.g., <user_id>/<app_id>
    description (str): model description

  """
  field_maps: FieldMapsConfig = None
  output_type: str = None
  labels: List[str] = field(default_factory=list)
  inference_parameters: List[InferParam] = field(default_factory=list)
  clarifai_model_id: str = ""
  type: str = ""
  clarifai_user_app_id: str = ""
  description: str = ""

  def _checking(self, var_name: str, var_value: Any):
    if var_name == "type":
      _model_types = MODEL_TYPES + [""]
      assert self.type in _model_types
    elif var_name == "clarifai_model_id" and var_value:
      # TODO: Must ensure name is valid
      pass
    elif var_name == "clarifai_user_app_id" and var_value:
      _user_app = var_value.split("/")
      assert len(_user_app) == 2, ValueError(
          f"id must be combination of user_id and app_id separated by `/`, e.g. <user_id>/<app_id>. Got {var_value}"
      )
    elif var_name == "labels":
      if var_value:
        assert isinstance(var_value, tuple) or isinstance(
            var_value, list), f"labels must be tuple or list, got {type(var_value)}"
        var_value = [str(each) for each in var_value]

    return var_value

  def __setattr__(self, __name: str, __value: Any) -> None:
    __value = self._checking(__name, __value)

    super().__setattr__(__name, __value)


@dataclass
class ModelConfigClass():
  """All config of model
  Args:
    clarifai_model (ClarifaiModelConfig): Clarifai model config
    serving_backend (ServingBackendConfig): Custom serving backend config. Only support triton for now
  """
  clarifai_model: ClarifaiModelConfig
  serving_backend: ServingBackendConfig

  def make_triton_model_config(
      self,
      model_name: str,
      model_version: str,
      image_shape: List = None,
      instance_group: Device = Device(),
      dynamic_batching: DynamicBatching = DynamicBatching(),
      max_batch_size: int = 1,
      backend: str = "python",
  ) -> TritonModelConfig:

    return TritonModelConfig(
        model_name=model_name,
        model_version=model_version,
        image_shape=image_shape,
        instance_group=instance_group,
        dynamic_batching=dynamic_batching,
        max_batch_size=max_batch_size,
        backend=backend,
        input=self.serving_backend.triton.input,
        output=self.serving_backend.triton.output)

  def dump_to_user_config(self):
    data = asdict(self)
    _self = deepcopy(self)
    # dump backend
    if hasattr(_self.serving_backend, "triton"):
      dict_triton_config = asdict(_self.serving_backend.triton)
      for k, v in dict_triton_config.items():
        if (k == "max_batch_size" and v > 1) \
        or (k == "image_shape" and v != [-1, -1] and self.clarifai_model.type in ModelTypes().image_input_models):
          continue
        else:
          data["serving_backend"]["triton"].pop(k, None)

      if not data["serving_backend"]["triton"]:
        data["serving_backend"].pop("triton", None)
    if not data["serving_backend"]:
      data.pop("serving_backend", None)

    # dump clarifai model
    data["clarifai_model"].pop("field_maps", None)
    data["clarifai_model"].pop("output_type", None)

    return data

  @classmethod
  def custom_doc(cls):
    msg = f"{cls.__doc__}\nWhere: \n\n"
    for k, v in cls.__annotations__.items():
      msg += f"* {k}:\n------\n {v.__doc__}\n"
    return msg


def read_yaml(path: str) -> dict:
  with open(path, encoding="utf-8") as f:
    config = yaml.safe_load(f)  # model dict
  return config


def parse_config(config: dict):
  clarifai_model = config.get("clarifai_model", {})
  serving_backend = config.get("serving_backend", {})
  if serving_backend:
    if serving_backend.get("triton", {}):
      # parse triton input/output
      triton = serving_backend["triton"]
      input_triton_configs = triton.pop("input", {})
      triton.update(
          dict(input=[
              InputConfig(
                  name=input["name"],
                  data_type=eval(f"DType.{input['data_type']}") if isinstance(
                      input['data_type'], str) else input['data_type'],
                  dims=input["dims"],
                  optional=input.get("optional", False),
              ) for input in input_triton_configs
          ]))
      output_triton_configs = triton.pop("output", {})
      triton.update(
          dict(output=[
              OutputConfig(
                  name=output["name"],
                  data_type=eval(f"DType.{output['data_type']}") if isinstance(
                      output['data_type'], str) else output['data_type'],
                  dims=output["dims"],
                  label_filename=output["label_filename"],
              ) for output in output_triton_configs
          ]))
      serving_backend.update(dict(triton=TritonModelConfig(**triton)))
    serving_backend = ServingBackendConfig(**serving_backend)

  # parse field maps for deployment
  field_maps = clarifai_model.pop("field_maps", {})
  clarifai_model.update(dict(field_maps=FieldMapsConfig(**field_maps)))
  # parse inference_parameters
  inference_parameters = clarifai_model.pop("inference_parameters", [])
  if inference_parameters is None:
    inference_parameters = []
  clarifai_model.update(
      dict(inference_parameters=[InferParam(**each) for each in inference_parameters]))
  # parse output type
  output_type = clarifai_model.pop("output_type", None)
  if output_type:
    #if isinstance(output_type, str):
    #  output_type = eval(output_type)
    clarifai_model.update(dict(output_type=output_type))

  clarifai_model = ClarifaiModelConfig(**clarifai_model)

  return ModelConfigClass(clarifai_model=clarifai_model, serving_backend=serving_backend)


def get_model_config(model_type: str) -> ModelConfigClass:
  """
  Get model config by model type

  Args:

    model_type (str): One of field value of ModelTypes

  Return:
    ModelConfigClass

  ### Example:
  >>> from clarifai.models.model_serving.model_config import get_model_config, ModelTypes
  >>> cfg = get_model_config(ModelTypes.text_classifier)
  >>> custom_triton_config = cfg.make_triton_model_config(**kwargs)

  """
  if model_type == "MODEL_TYPE_PLACEHOLDER":
    logger.warning(
        "Warning: A placeholder value has been detected for obtaining the model configuration. This will result in empty `ModelConfigClass` object."
    )
    return ModelConfigClass(clarifai_model=None, serving_backend=None)

  import os
  assert model_type in MODEL_TYPES, f"`model_type` must be in {MODEL_TYPES}"
  cfg = read_yaml(
      os.path.join(os.path.dirname(__file__), "model_types_config", f"{model_type}.yaml"))
  cfg = parse_config(cfg)
  cfg.clarifai_model.type = model_type
  return cfg


_model_types = ModelTypes()
MODEL_TYPES = _model_types.all
del _model_types


def load_user_config(cfg_path: str) -> ModelConfigClass:
  """Read `clarifai_config.yaml` in user working dir

  Args:
      cfg_path (str): path to config

  Returns:
      ModelConfigClass
  """
  cfg = read_yaml(cfg_path)
  return _ensure_user_config(cfg)


def _ensure_user_config(user_config: dict) -> ModelConfigClass:
  """Ensure user config with default one

  Args:
      user_config (dict): ModelConfigClass as dict

  Raises:
      e: Exception when loading user config

  Returns:
      ModelConfigClass
  """

  try:
    user_config_obj: ModelConfigClass = parse_config(user_config)
  except Exception as e:
    raise e

  default_config = get_model_config(user_config_obj.clarifai_model.type)

  for _model_cfg, value in asdict(user_config_obj.clarifai_model).items():

    if value and _model_cfg != "field_maps":
      setattr(default_config.clarifai_model, _model_cfg, value)

  if hasattr(user_config_obj, "serving_backend"):
    if hasattr(user_config_obj.serving_backend, "triton"):
      if user_config_obj.serving_backend.triton.max_batch_size > 1:
        default_config.serving_backend.triton.max_batch_size = user_config_obj.serving_backend.triton.max_batch_size
      if user_config_obj.serving_backend.triton.image_shape != [-1, -1]:
        default_config.serving_backend.triton.image_shape = user_config_obj.serving_backend.triton.image_shape

  return default_config



from argparse import Namespace, _SubParsersAction

from ._utils import list_model_upload_examples
from .base import BaseClarifaiCli


class ExampleCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: _SubParsersAction):
    creator_parser = parser.add_parser("example", help="Download/List examples of model upload")
    sub_creator_parser = creator_parser.add_subparsers()

    SubListExampleCli.register(sub_creator_parser)

    creator_parser.set_defaults(func=ExampleCli)


class SubListExampleCli(BaseClarifaiCli):

  @staticmethod
  def register(parser: _SubParsersAction):
    _parser = parser.add_parser("list")
    _parser.add_argument("--force-download", action="store_true", help="Force download examples")
    _parser.set_defaults(func=SubListExampleCli)

  def __init__(self, args: Namespace) -> None:
    self.force_download = args.force_download

  def run(self):
    _list = list_model_upload_examples(self.force_download)
    print(f"Found {len(_list)} examples")
    for each in _list:
      print(f" * {each}")



import os
from copy import deepcopy
from typing import Dict, Iterable, List, Union

import numpy as np
import yaml

from clarifai.models.model_serving.constants import IMAGE_TENSOR_NAME, TEXT_TENSOR_NAME
from clarifai.models.model_serving.model_config import (
    ClassifierOutput, EmbeddingOutput, ImageOutput, InferParam, InferParamManager, MasksOutput,
    ModelTypes, TextOutput, VisualDetector, load_user_config)

_default_texts = ["Photo of a cat", "A cat is playing around", "Hello, this is test"]

_default_images = [
    np.zeros((100, 100, 3), dtype='uint8'),  #black
    np.ones((100, 100, 3), dtype='uint8') * 255,  #white
    np.random.uniform(0, 255, (100, 100, 3)).astype('uint8')  #noise
]


def _is_valid_logit(x: np.array):
  return np.all(0 <= x) and np.all(x <= 1)


def _is_non_negative(x: np.array):
  return np.all(x >= 0)


def _is_integer(x):
  return np.all(np.equal(np.mod(x, 1), 0))


class BaseTest:
  init_inference_parameters = {}

  def __init__(self, init_inference_parameters={}) -> None:
    import sys
    if 'inference' in sys.modules:
      del sys.modules['inference']
    import inference
    from inference import InferenceModel
    self.model = InferenceModel()
    self._base_dir = os.path.dirname(inference.__file__)
    self.cfg_path = os.path.join(self._base_dir, "clarifai_config.yaml")
    self.user_config = load_user_config(self.cfg_path)
    self._user_labels = None
    # check if labels exists
    for output_config in self.user_config.serving_backend.triton.output:
      if output_config.label_filename:
        self._user_labels = self.user_config.clarifai_model.labels
        assert self._user_labels, f"Model type `{self.user_config.clarifai_model.type}` requires labels, "\
        f"but can not found value of `clarifai_model.labels` in {self.cfg_path}. Please update this attribute to build the model"

    # update init vs user_defined params
    user_defined_infer_params = [
        InferParam(**each) for each in self.user_config.clarifai_model.inference_parameters
    ]
    total_infer_params = []
    if init_inference_parameters:
      self.init_inference_parameters = init_inference_parameters
    for k, v in self.init_inference_parameters.items():
      _exist = False
      for user_param in user_defined_infer_params:
        if user_param.path == k:
          if user_param.default_value != v:
            print(f"Warning: Overwrite parameter `{k}` with default value `{v}`")
          user_param.default_value = v
          _exist = True
          total_infer_params.append(user_param)
          user_defined_infer_params.remove(user_param)
          break
      if not _exist:
        total_infer_params.append(InferParamManager.from_kwargs(**{k: v}).params[0])

    self.infer_param_manager = InferParamManager(
        params=total_infer_params + user_defined_infer_params)
    self.user_config.clarifai_model.inference_parameters = self.infer_param_manager.get_list_params(
    )
    self._overwrite_cfg()

  @property
  def user_labels(self):
    return self._user_labels

  def _overwrite_cfg(self):
    config = yaml.dump(self.user_config.dump_to_user_config(),)
    with open(self.cfg_path, "w") as f:
      f.write(config)

  def predict(self, input_data: Union[List[np.ndarray], List[str], Dict[str, Union[List[
      np.ndarray], List[str]]]], **inference_parameters) -> Iterable:
    """
    Test Prediction method is exact `InferenceModel.predict` method with
    checking inference paramters.

    Args:
    -----
    - input_data: A list of input data item to predict on. The type depends on model input type:
      * `image`: List[np.ndarray]
      * `text`: List[str]
      * `multimodal`:
        input_data is list of dict where key is input type name e.i. `image`, `text` and value is list.
        {"image": List[np.ndarray], "text": List[str]}

    - **inference_parameters: keyword args of your inference parameters.

    Returns:
    --------
      List of your inference model output type
    """
    infer_params = self.infer_param_manager.validate(**inference_parameters)
    outputs = self.model.predict(input_data=input_data, inference_parameters=infer_params)
    outputs = self._verify_outputs(outputs)
    return outputs

  def _verify_outputs(self, outputs: List[Union[ClassifierOutput, VisualDetector, EmbeddingOutput,
                                                TextOutput, ImageOutput, MasksOutput]]):
    """Test output value/dims

    Args:
      outputs (List[Union[ClassifierOutput, VisualDetector, EmbeddingOutput, TextOutput, ImageOutput, MasksOutput]]): Outputs of `predict` method
    """
    _outputs = deepcopy(outputs)
    _output = _outputs[0]

    if isinstance(_output, EmbeddingOutput):
      # not test
      pass
    elif isinstance(_output, ClassifierOutput):
      for each in _outputs:
        assert _is_valid_logit(each.predicted_scores), "`predicted_scores` must be in range [0, 1]"
        assert len(each.predicted_scores) == len(
            self.user_labels
        ), f"`predicted_scores` dim must be equal to labels, got {len(each.predicted_scores)} != labels {len(self.user_labels)}"
    elif isinstance(_output, VisualDetector):
      for each in _outputs:
        assert _is_valid_logit(each.predicted_scores), "`predicted_scores` must be in range [0, 1]"
        assert _is_integer(each.predicted_labels), "`predicted_labels` must be integer"
        assert np.all(0 <= each.predicted_labels) and np.all(each.predicted_labels < len(
            self.user_labels)), f"`predicted_labels` must be in [0, {len(self.user_labels) - 1}]"
        assert _is_non_negative(each.predicted_bboxes), "`predicted_bboxes` must be >= 0"
    elif isinstance(_output, MasksOutput):
      for each in _outputs:
        assert np.all(0 <= each.predicted_mask) and np.all(each.predicted_mask < len(
            self.user_labels)), f"`predicted_mask` must be in [0, {len(self.user_labels) - 1}]"
    elif isinstance(_output, TextOutput):
      pass
    elif isinstance(_output, ImageOutput):
      for each in _outputs:
        assert _is_non_negative(each.image), "`image` must be >= 0"
    else:
      pass

    return outputs

  def test_with_default_inputs(self):
    model_type = self.user_config.clarifai_model.type
    if model_type == ModelTypes.multimodal_embedder:
      self.predict(input_data=[{IMAGE_TENSOR_NAME: each} for each in _default_images])
      self.predict(input_data=[{TEXT_TENSOR_NAME: each} for each in _default_texts])
      self.predict(input_data=[{
          TEXT_TENSOR_NAME: text,
          IMAGE_TENSOR_NAME: img
      } for text, img in zip(_default_texts, _default_images)])
    elif model_type.startswith("visual"):
      self.predict(input_data=_default_images)
    else:
      self.predict(input_data=_default_texts)
